@@ -1,13 +1,13 @@
 from collections import namedtuple
 from datetime import date, datetime
-from decimal import Decimal
+from decimal import Decimal, InvalidOperation
 import re
 
 from lxml import etree
 from stream_unzip import UnzipValueError, stream_unzip
 
 
-def stream_read_ods(ods_chunks, chunk_size=65536):
+def stream_read_ods(ods_chunks, max_string_length=65536, max_columns=65536, chunk_size=65536):
 
     # lxml iterparse takes a file-like object, but stream_read_ods accepts an iterable
     # so we have to do some low-ish level faffing to convert from one to the other
@@ -73,11 +73,22 @@ def stream_read_ods(ods_chunks, chunk_size=65536):
         ns_table = '{urn:oasis:names:tc:opendocument:xmlns:table:1.0}'
         ns_office = '{urn:oasis:names:tc:opendocument:xmlns:office:1.0}'
 
+        def _next(it):
+            try:
+                return next(it)
+            except etree.LxmlError as e:
+                raise InvalidContentXMLError() from e
+
+        def _append(l, value):
+            if len(l) == max_columns:
+                raise TooManyColumnsError(max_columns)
+            l.append(value)
+
         def table_rows(parsed_xml_it):
             row = None
 
             while True:
-                event, element = next(parsed_xml_it)
+                event, element = _next(parsed_xml_it)
 
                 # Starting a row
                 if event == 'start' and f'{ns_table}table-row' == element.tag:
@@ -89,7 +100,7 @@ def stream_read_ods(ods_chunks, chunk_size=65536):
 
                 # Starting a table cell
                 if event == 'start' and f'{ns_table}table-cell' == element.tag:
-                    row.append(table_cell(parsed_xml_it, element))
+                    _append(row, table_cell(parsed_xml_it, element))
 
                 # Ending the table
                 if event == 'end' and f'{ns_table}table' == element.tag:
@@ -99,58 +110,106 @@ def stream_read_ods(ods_chunks, chunk_size=65536):
 
         def table_cell(parsed_xml_it, cell_element):
             value_type = cell_element.attrib.get(f'{ns_office}value-type')
+            return \
+                None if value_type is None else \
+                parse_boolean(cell_element) if value_type == 'boolean' else \
+                parse_currency(cell_element) if value_type == 'currency' else \
+                parse_date(cell_element) if value_type == 'date' else \
+                parse_float(cell_element) if value_type == 'float' else \
+                parse_percentage(cell_element) if value_type == 'percentage' else \
+                parse_string(cell_element, parsed_xml_it) if value_type == 'string' else \
+                parse_time(cell_element) if value_type == 'time' else \
+                error(InvalidTypeError(value_type))
 
-            # Non-strings are always from attributes
-            if value_type != 'string':
-                return \
-                    None if value_type is None else \
-                    parse_boolean(cell_element) if value_type == 'boolean' else \
-                    parse_currency(cell_element) if value_type == 'currency' else \
-                    parse_date(cell_element) if value_type == 'date' else \
-                    parse_float(cell_element) if value_type == 'float' else \
-                    parse_percentage(cell_element) if value_type == 'percentage' else \
-                    parse_time(cell_element) if value_type == 'time' else \
-                    value_error(value_type)
-
-            # Strings can be from an attribute...
-            attribute_string_value = cell_element.attrib.get(f'{ns_office}:string-value')
-            if attribute_string_value is not None:
-                return attribute_string_value
-
-            # ... but otherwise extract from contents
-            while True:
-                event, element = next(parsed_xml_it)
-                if event == 'end' and element is cell_element:
-                    return ''.join(cell_element.itertext())
-
-        def value_error(message):
-            raise ValueError(message)
+        def error(e):
+            raise e(message)
 
         def parse_boolean(cell_element):
             value = cell_element.attrib[f'{ns_office}boolean-value']
             return \
                 True if value == 'true' else \
                 False if value == 'false' else \
-                value_error(value)
+                error(InvalidBooleanValueError(value))
 
         def parse_currency(cell_element):
-            return Currency(cell_element.attrib[f'{ns_office}value'], code=cell_element.attrib.get(f'{ns_office}currency'))
+            value = cell_element.attrib[f'{ns_office}value']
+            try:
+                return Currency(value, code=cell_element.attrib.get(f'{ns_office}currency'))
+            except InvalidOperation as e:
+                raise InvalidCurrencyValueError(value) from e
 
         def parse_date(cell_element):
             value = cell_element.attrib[f'{ns_office}date-value']
             try:
-                return date.fromisoformat(value)
-            except ValueError:
-                return datetime.fromisoformat(value)
+                try:
+                    return date.fromisoformat(value)
+                except ValueError:
+                    return datetime.fromisoformat(value)
+            except ValueError as e:
+                raise InvalidDateValueError(value) from e
 
         def parse_float(cell_element):
-            return Decimal(cell_element.attrib[f'{ns_office}value'])
+            value = cell_element.attrib[f'{ns_office}value']
+            try:
+                return Decimal(value)
+            except InvalidOperation as e:
+                raise InvalidFloatValueError(value) from e
 
         def parse_percentage(cell_element):
-            return Percentage(cell_element.attrib[f'{ns_office}value'])
+            value = cell_element.attrib[f'{ns_office}value']
+            try:
+                return Percentage(value)
+            except InvalidOperation as e:
+                raise InvalidPercentageValueError(e) from e
+
+        def parse_string(cell_element, parsed_xml_it):
+
+            def itertext():
+                # Like lxml's itertext, but doesn't use recursion and clears memory along the way
+                l = 0
+                previous_event = None
+                stack = ['', '']
+                while True:
+                    event, element = _next(parsed_xml_it)
+
+                    if event == 'start' and previous_event == 'start':
+                        stack.append('')
+                        continue
+
+                    element_text = (element.text or '')
+                    element_tail = (element.tail or '') if element is not cell_element else ''
+                    l += len(element_text) + len(element_tail)
+                    if l > max_string_length:
+                        raise StringTooLongError(max_string_length)
+
+                    if event == 'end' and previous_event == 'end':
+                        popped = stack.pop()
+                        stack[-1] += element_text + popped + element_tail
+
+                    if event == 'end' and previous_event != 'end':
+                        stack[-1] += element_text + element_tail
+
+                    clear_mem(event, element)
+
+                    if event == 'end' and element is cell_element:
+                        break
+
+                    previous_event = event
+
+                return stack[0]
+
+            attribute_string_value = cell_element.attrib.get(f'{ns_office}:string-value')
+            return \
+                attribute_string_value if attribute_string_value is not None else \
+                itertext()
 
         def parse_time(cell_element):
-            time_dict = time_regex.match(cell_element.attrib[f'{ns_office}time-value']).groupdict(0)
+            value = cell_element.attrib[f'{ns_office}time-value']
+            try:
+                time_dict = time_regex.match(value).groupdict(0)
+            except AttributeError as e:
+                raise InvalidTimeValueError(value) from e
+
             return Time(
                 time_dict.get('sign') or '+',
                 int(time_dict.get('years', '0')),
@@ -172,7 +231,7 @@ def stream_read_ods(ods_chunks, chunk_size=65536):
 
         while True:
             try:
-                event, element = next(parsed_xml_it)
+                event, element = _next(parsed_xml_it)
             except StopIteration:
                 break
 
@@ -189,14 +248,12 @@ def stream_read_ods(ods_chunks, chunk_size=65536):
     unzipped_member_files = stream_unzip(ods_chunks, chunk_size=chunk_size)
     content_xml_chunks = validate_mimetype_and_get_content(unzipped_member_files)
     content_xml_file_like_obj = to_file_like_obj(content_xml_chunks)
-    content_xml_parsed = etree.iterparse(content_xml_file_like_obj, events=('start', 'end'))
+    content_xml_parsed = etree.iterparse(content_xml_file_like_obj, events=('start', 'end'), resolve_entities=False)
 
     try:
         yield from get_sheets_and_rows(content_xml_parsed)
     except UnzipValueError as e:
         raise UnzipError() from e
-    except etree.LxmlError as e:
-        raise InvalidContentXMLError() from e
 
 
 def simple_table(rows, skip_rows=0):
@@ -276,3 +333,47 @@ class MissingContentXMLError(InvalidODSFileError):
 
 class InvalidContentXMLError(InvalidODSFileError):
     pass
+
+
+class InvalidODSXMLError(InvalidODSFileError):
+    pass
+
+
+class InvalidTypeError(InvalidODSXMLError):
+    pass
+
+
+class InvalidValueError(InvalidODSXMLError):
+    pass
+
+
+class InvalidBooleanValueError(InvalidValueError):
+    pass
+
+
+class InvalidDateValueError(InvalidValueError):
+    pass
+
+
+class InvalidFloatValueError(InvalidValueError):
+    pass
+
+
+class InvalidPercentageValueError(InvalidValueError):
+    pass
+
+
+class InvalidTimeValueError(InvalidValueError):
+    pass
+
+
+class SizeError(StreamReadODSError):
+    pass
+
+
+class TooManyColumnsError(StreamReadODSError):
+    pass
+
+
+class StringTooLongError(StreamReadODSError):
+    pass
